use std::collections::HashMap;

use camino::Utf8PathBuf;

use crate::{common::Loc, token::Token};

#[derive(Debug, Eq, PartialEq, Clone, Hash)]
pub struct Ident {
    pub text: String,
    pub loc: Loc,
}

impl Ident {
    pub fn from_tok(t: Token) -> Self {
        let text;
        match t.typ {
            crate::token::TokenType::Ident(s) => text = s,
            _ => unreachable!(),
        }
        Self { text, loc: t.loc.clone() }
    }
}

#[derive(Debug, Clone)]
pub struct Type {
    pub name: Ident,
    pub inner: Option<Box<Type>>,
}

#[derive(Debug, Clone)]
pub struct Program {
    pub file: Utf8PathBuf,
    pub items: Vec<Stat>
}

#[derive(Debug, Clone)]
pub struct Stat {
    pub typ: StatType,
    pub loc: Loc,
}

#[derive(Debug, Clone)]
pub enum StatType {
    Struct(StructStat),
    Enum(EnumStat),
    Var(VarStat),
    Set(SetStat),
    Match(MatchStat), 
    Not(UnaryStat),
    Eq(BinaryStat),
    Neq(BinaryStat),
    Lt(BinaryStat),
    Gt(BinaryStat),
    Le(BinaryStat),
    Ge(BinaryStat),
    Add(BinaryStat),
    Sub(BinaryStat),
    Mul(BinaryStat),
    Div(BinaryStat),
    Mod(BinaryStat),
    Shr(BinaryStat),
    Shl(BinaryStat),
    Band(BinaryStat),
    Bor(BinaryStat),
    Bnot(UnaryStat),
    And(BinaryStat),
    Or(BinaryStat),
    Xor(BinaryStat),
    Inc(UnaryStat),
    Dec(UnaryStat),
    Ref(UnaryStat),
    Export(ExportStat),
    Function(FnStat),
    FunctionCall(FnCallStat),
    Return(ReturnStat),
    If(IfStat),
    While(WhileStat),
    Block(BlockStat),
    Literal(ValueType)
}

#[derive(Debug, Clone)]
pub enum ValueType {
    Float(f64),
    Int(i64),
    Bool(bool),
    String(String),
    Char(char),
}

#[derive(Debug, Clone)]
pub struct BlockStat(pub Vec<Stat>);
#[derive(Debug, Clone)]
pub struct UnaryStat(pub Box<Stat>);
#[derive(Debug, Clone)]
pub struct BinaryStat(pub Box<Stat>, pub Box<Stat>);

#[derive(Debug, Clone)]
pub struct MatchStat {
    pub val: Ident,
    pub cases: HashMap<ValueType, BlockStat>
}

#[derive(Debug, Clone)]
pub struct SetStat {
    pub name: Ident,
    pub val: Box<Stat>,
}

#[derive(Debug, Clone)]
pub struct VarStat {
    pub name: Ident,
    pub typ: Option<Type>,
    pub val: Option<Box<Stat>>,
}

#[derive(Debug, Clone)]
pub struct EnumStat {
    pub name: Ident,
    pub fields: Vec<Ident>
}

#[derive(Debug, Clone)]
pub struct StructStat {
    pub name: Ident,
    pub fields: HashMap<Ident, Type>,
}

#[derive(Debug, Clone)]
pub struct FnStat {
    pub name: Ident,
    pub args: HashMap<Ident, Type>,
    pub body: BlockStat,
}

#[derive(Debug, Clone)]
pub struct FnCallStat {
    pub name: Ident,
    pub args: Vec<Ident>,
}

#[derive(Debug, Clone)]
pub struct ReturnStat {
    pub arg: Option<Box<Stat>>,
}

#[derive(Debug, Clone)]
pub struct IfStat {
    pub condition: Box<Stat>,
    pub body: BlockStat,
    pub els: Box<Stat>
}

#[derive(Debug, Clone)]
pub struct WhileStat {
    pub condition: Box<Stat>,
    pub body: BlockStat,
}

#[derive(Debug, Clone)]
pub struct ExportStat {
    pub name: Ident,
}

