use std::collections::HashMap;

use anyhow::{bail, Result};
use camino::Utf8Path;

use crate::{ast::{self, StructStat}, common::Loc, token::{KeywordType, Token, TokenType}};


pub struct Parser {
    tokens: Vec<Token>,
    loc: Loc,
    is_in_func: bool,
}

impl Parser {
    pub fn new() -> Self {
        Self {
            tokens: Vec::new(),
            loc: Loc::new(""),
            is_in_func: false,
        }
    }

    pub fn parse(&mut self, file: &Utf8Path, _tokens: &Vec<Token>) -> Result<ast::Program> {
        self.tokens = _tokens.to_vec();
        self.tokens.reverse();
        let mut prog = ast::Program {
            file: file.into(),
            items: Vec::new()
        };

        while !self.tokens.is_empty() {
            prog.items.push(self.parse_stat()?);
        }

        Ok(prog)
    }

    fn parse_stat(&mut self) -> Result<ast::Stat> {
        log::debug!("parse_stat");
        self.expect_next(TokenType::ParenL)?;
        log::debug!("parse_stat.2");
        let Some(main) = self.tokens.pop() else {
            log::error!("Expected function, struct, enum definitions, but found nothing");
            bail!("");
        };
        self.loc = main.loc;
        let ret = match main.typ {
            TokenType::ParenL => {
                ast::Stat {
                    loc: self.loc.clone(),
                    typ: ast::StatType::Block(self.parse_block_stat()?),
                }
            }
            TokenType::Keyword(kw) => {
                log::debug!("kw: {kw:?}");
                match kw {
                    KeywordType::Eq   | KeywordType::Neq  |
                    KeywordType::Lt   | KeywordType::Gt   |
                    KeywordType::Ge   | KeywordType::Le   |
                    KeywordType::Or   | KeywordType::And  |
                    KeywordType::Shr  | KeywordType::Shl  |
                    KeywordType::Bor  | KeywordType::Band |
                    KeywordType::Xor  | KeywordType::Add  |
                    KeywordType::Sub  | KeywordType::Div  |
                    KeywordType::Mod  | KeywordType::Mul => {
                        if !self.is_in_func {
                            log::error!("Can only use {:?} in a function", kw);
                            bail!("");
                        }
                        self.parse_binary_stat(kw)?
                    }

                    KeywordType::Not | KeywordType::Bnot |
                    KeywordType::Inc | KeywordType::Dec  |
                    KeywordType::Ref => {
                        if !self.is_in_func {
                            log::error!("Can only use {:?} in a function", kw);
                            bail!("");
                        }
                        self.parse_unary_stat(kw)?
                    }
                    KeywordType::Var | KeywordType::Set |
                    KeywordType::If  | KeywordType::While |
                    KeywordType::Match  if !self.is_in_func => {
                        log::error!("Can only use {:?} in a function", kw);
                        bail!("");
                    }
                    // TODO: Dont allow structs exports, and enums in functions
                    KeywordType::Struct => self.parse_struct_stat()?,
                    KeywordType::Fn => self.parse_fn_stat()?,
                    KeywordType::Export => self.parse_export_stat()?,
                    KeywordType::Enum => self.parse_enum_stat()?,
                    
                    KeywordType::Call if self.is_in_func => self.parse_call_stat()?,
                    KeywordType::Var if self.is_in_func => self.parse_var_stat()?,
                    KeywordType::Set if self.is_in_func => self.parse_set_stat()?,
                    KeywordType::If if self.is_in_func => self.parse_if_stat()?,
                    KeywordType::While if self.is_in_func => self.parse_while_stat()?,
                    KeywordType::Match if self.is_in_func => self.parse_match_stat()?,
                    KeywordType::Return if self.is_in_func => self.parse_return_stat()?,

                    KeywordType::Else => {
                        log::error!("Keyword else is only suppost to be put after an else statement");
                        bail!("");
                    }
                    v => unreachable!("{v:?}")
                }
            }
            _ => {
                log::error!("Expected function, struct, enum definitions, but found {:?}", main.typ);
                bail!("");
            }
        };
        Ok(ret)
    }
    
    

    fn parse_binary_stat(&mut self, kw: KeywordType) -> Result<ast::Stat> {
        let bs = ast::BinaryStat(
            Box::new(self.parse_stat()?), 
            Box::new(self.parse_stat()?)
        );

        let typ = match kw {
            KeywordType::Eq => ast::StatType::Eq(bs),
            KeywordType::Neq => ast::StatType::Neq(bs),
            KeywordType::Lt => ast::StatType::Lt(bs),
            KeywordType::Gt => ast::StatType::Gt(bs),
            KeywordType::Ge => ast::StatType::Ge(bs),
            KeywordType::Le => ast::StatType::Le(bs),
            KeywordType::Or => ast::StatType::Or(bs),
            KeywordType::And  => ast::StatType::And(bs),
            KeywordType::Add  => ast::StatType::Add(bs),
            KeywordType::Sub  => ast::StatType::Sub(bs),
            KeywordType::Div  => ast::StatType::Div(bs),
            KeywordType::Mul  => ast::StatType::Mul(bs),
            KeywordType::Mod  => ast::StatType::Mod(bs),
            KeywordType::Shr => ast::StatType::Shr(bs),
            KeywordType::Shl => ast::StatType::Shl(bs),
            KeywordType::Bor => ast::StatType::Bor(bs),
            KeywordType::Band => ast::StatType::Band(bs),
            KeywordType::Xor => ast::StatType::Xor(bs),
            _ => unreachable!("{kw:?}")
        };
        self.expect_next(TokenType::ParenR)?;
        Ok(ast::Stat {
            loc: self.loc.clone(),
            typ,
        })
    }

    fn parse_unary_stat(&mut self, kw: KeywordType) -> Result<ast::Stat> {
        let us = ast::UnaryStat(Box::new(self.parse_stat()?));
        let typ = match kw {
            KeywordType::Not  => ast::StatType::Not(us),
            KeywordType::Bnot => ast::StatType::Bnot(us),
            KeywordType::Inc  => ast::StatType::Inc(us),
            KeywordType::Dec  => ast::StatType::Dec(us),
            KeywordType::Ref  => ast::StatType::Ref(us),
            _ => unreachable!("{kw:?}")
        };
        self.expect_next(TokenType::ParenR)?;
        Ok(ast::Stat { typ, loc: self.loc.clone() })
    }

    fn parse_block_stat(&mut self) -> Result<ast::BlockStat> {
        log::debug!("parse_block_stat.1");
        self.expect_next(TokenType::ParenL)?;
        let mut bs = Vec::new();
        while let Some(t) = self.tokens.last() {
            match &t.typ {
                TokenType::ParenL => {
                    log::debug!("parse_block_stat.2");
                    bs.push(self.parse_stat()?);
                }
                TokenType::ParenR => {
                    log::debug!("parse_block_stat.3");
                    self.expect_next(TokenType::ParenR)?;
                    break;
                }
                typ => {
                    log::debug!("parse_block_stat.4");
                    log::error!("{}: Expected '(' found {typ:?}", self.loc);
                    bail!("");
                }
            }
        }

        Ok(ast::BlockStat(bs))
    }
    fn parse_struct_stat(&mut self) -> Result<ast::Stat> {
        todo!();
    }

    fn parse_fn_stat(&mut self) -> Result<ast::Stat> {
        self.is_in_func = true;
        let name = self.expect_next(TokenType::Ident(String::new()))?;
        let mut args: HashMap<ast::Ident, ast::Type> = HashMap::new();

        log::debug!("parse_fn_stat.1");
        self.expect_next(TokenType::ParenL)?;
        log::debug!("parse_fn_stat.2");

        while self.try_next(TokenType::ParenR).is_none() {
            log::debug!("parse_fn_stat.3");
            self.expect_next(TokenType::ParenL)?;
            log::debug!("parse_fn_stat.4");
            let arg_name = self.expect_next(TokenType::Ident(String::new()))?;
            let typ = self.parse_type()?;
            self.expect_next(TokenType::ParenR)?;
            args.insert(ast::Ident::from_tok(arg_name), typ);
        }

        log::debug!("parse_fn_stat.5");
        self.expect_next(TokenType::ParenR)?;

        let body = self.parse_block_stat()?;
        log::debug!("parse_fn_stat.6");

        let fs = ast::FnStat {
            name: ast::Ident::from_tok(name.clone()),
            args,
            body
        };
        
        self.is_in_func = false;

        self.expect_next(TokenType::ParenR)?;
        Ok(ast::Stat {
            loc: name.loc.clone(),
            typ: ast::StatType::Function(fs)
        })
    }

    fn parse_type(&mut self) -> Result<ast::Type> { // (args (arr string_t))
        let has_another = self.try_next(TokenType::ParenL).is_some();
        if has_another {
            self.tokens.pop();
        }
        let name = self.expect_next(TokenType::Ident(String::new()))?;
        let mut typ = ast::Type {
            name: ast::Ident::from_tok(name),
            inner: None,
        };

        if has_another {
            typ.inner = Some(Box::new(self.parse_type()?));
            self.expect_next(TokenType::ParenR)?;
        }
        Ok(typ)
   }


    fn parse_enum_stat(&mut self) -> Result<ast::Stat> {
        todo!();
    }
    fn parse_var_stat(&mut self) -> Result<ast::Stat> {
        let name = self.expect_next(TokenType::Ident(String::new()))?;
        let typ;
        let typ_tok = self.expect_several(&[
            TokenType::Ident(String::new()),
            TokenType::Underscore
        ])?;
        match &typ_tok.typ {
            TokenType::Underscore => {
                typ = None;
            }
            TokenType::Ident(_) => {
                self.tokens.push(typ_tok);
                typ = Some(self.parse_type()?);
            }
            _ => unreachable!()
        }
        let val;
        if self.try_next(TokenType::ParenR).is_some() {
            val = None;
        } else {
            val = Some(Box::new(self.parse_val()?));
        }
        
        self.expect_next(TokenType::ParenR)?;

        Ok(ast::Stat { 
            loc: name.loc.clone(),
            typ: ast::StatType::Var(ast::VarStat {
                name: ast::Ident::from_tok(name),
                typ,
                val,
            })
        })
    }

    fn parse_val(&mut self) -> Result<ast::Stat> {
        if self.try_next(TokenType::ParenL).is_some() {
            Ok(self.parse_stat()?)
        } else {
            let tok = self.expect_several(&[
                TokenType::String(String::new()),
                TokenType::Int(0),
                TokenType::Float(0.0),
                TokenType::Char('\0'),
            ])?;
            let vt = match tok.typ {
                TokenType::String(s) => ast::ValueType::String(s),
                TokenType::Int(i) => ast::ValueType::Int(i),
                TokenType::Float(f) => ast::ValueType::Float(f),
                TokenType::Char(c) => ast::ValueType::Char(c),
                _ => unreachable!("")
            };
            Ok(ast::Stat {
                loc: tok.loc.clone(),
                typ: ast::StatType::Literal(vt)
            })
        }
    }
    fn parse_set_stat(&mut self) -> Result<ast::Stat> {
        let name = self.expect_next(TokenType::Ident(String::new()))?;
        let val = self.parse_val()?;
        self.expect_next(TokenType::ParenR)?;

        Ok(ast::Stat { 
            loc: name.loc.clone(),
            typ: ast::StatType::Set(ast::SetStat {
                name: ast::Ident::from_tok(name),
                val: Box::new(val),
            })
        })
    }

    fn parse_call_stat(&mut self) -> Result<ast::Stat> {
        let name = self.expect_next(TokenType::Ident(String::new()))?;
        let mut args = Vec::new();

        while self.try_next(TokenType::ParenR).is_none() {
            args.push(ast::Ident::from_tok(self.expect_next(TokenType::Ident(String::new()))?));
        }
        self.expect_next(TokenType::ParenR)?;

        Ok(ast::Stat {
            loc: name.loc.clone(),
            typ: ast::StatType::FunctionCall(ast::FnCallStat {
                name: ast::Ident::from_tok(name), 
                args 
            })
        })
    }
    fn parse_return_stat(&mut self) -> Result<ast::Stat> {
        let loc = self.loc.clone();
        let val;
        if self.try_next(TokenType::ParenR).is_some() {
           val = None; 
        } else {
            val = Some(Box::new(self.parse_val()?));
        }
        self.expect_next(TokenType::ParenR)?;
        Ok(ast::Stat { 
            loc,
            typ: ast::StatType::Return(ast::ReturnStat {
                arg: val,
            })
        })
    }
    fn parse_if_stat(&mut self) -> Result<ast::Stat> {
        todo!();
    }
    fn parse_while_stat(&mut self) -> Result<ast::Stat> {
        todo!();
    }
    fn parse_match_stat(&mut self) -> Result<ast::Stat> {
        todo!();
    }
    fn parse_export_stat(&mut self) -> Result<ast::Stat> {
        todo!();
    }


    fn expect_next(&mut self, tt: TokenType) -> Result<Token> {
        if let Some(t) = self.tokens.pop() {
            log::debug!("en: {:?}", t.typ);
            if std::mem::discriminant(&t.typ) == std::mem::discriminant(&tt) {
                self.loc = t.loc.clone();
                return Ok(t);
            } else {
                log::error!("{}: Expected {:?}, got {:?}", t.loc, tt, t.typ);
                bail!("")
            }
        }
        log::error!("{}: Expected {:?}, got Nothing", self.loc, tt);
        bail!("")
    }

    fn expect_several(&mut self, tt: &[TokenType]) -> Result<Token> {
        let tt_text = tt
            .into_iter()
            .map(|v| format!("{:?}", v))
            .collect::<Vec<String>>()
            .join(", ");
        
        if let Some(t) = self.tokens.pop() {
            for typ in tt {
                if std::mem::discriminant(&t.typ) == std::mem::discriminant(&typ) {
                    self.loc = t.loc.clone();
                    return Ok(t);
                }
            }
            log::error!("{}: Expected {}, got {:?}", t.loc, tt_text, t.typ);
            bail!("")
        }
        log::error!("{}: Expected {}, got Nothing", self.loc, tt_text);
        bail!("")
    }
    fn try_next(&mut self, tt: TokenType) -> Option<&Token> {
        if let Some(t) = self.tokens.last() {
            if std::mem::discriminant(&t.typ) == std::mem::discriminant(&tt) {
                return Some(t);
            }
        }
        None
    }
}

