use anyhow::{anyhow, bail};

use crate::common::Loc;


#[derive(Debug, Clone)]
pub struct Token {
    pub typ: TokenType,
    pub loc: Loc,
}

impl Token {
    pub fn new(typ: TokenType, loc: Loc) -> Self {
        Self {typ, loc}
    }
}

#[derive(Debug, Clone, PartialEq, PartialOrd)]
pub enum TokenType {
    Ident(String),
    Keyword(KeywordType),
    Int(i64),
    Float(f64),
    String(String),
    Char(char),
    Underscore,
    ParenR,
    ParenL,
    Dot,
}

#[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord)]
pub enum KeywordType {
    Struct,
    Enum,
    Fn,
    Var,
    Set,
    If,
    While,
    Match, 
    Not,
    Eq,
    Neq,
    Lt,
    Gt,
    Le,
    Ge,
    Add,
    Sub,
    Div,
    Mod,
    Mul,
    Shr,
    Shl,
    Band,
    Bor,
    Bnot,
    And,
    Or,
    Xor,
    Inc,
    Dec,
    Ref,
    Export,
    Else,
    Call,
    Return,
    StructInit,
}

impl TryFrom<&str> for KeywordType {
    type Error = anyhow::Error;
    fn try_from(value: &str) -> Result<Self, Self::Error> {
        match value {
            "fn" => Ok(Self::Fn),
            "var" => Ok(Self::Var),
            "set" => Ok(Self::Set),
            "if" => Ok(Self::If),
            "while" => Ok(Self::While),
            "match" => Ok(Self::Match),
            "not" => Ok(Self::Not),
            "eq" => Ok(Self::Eq),
            "neq" => Ok(Self::Neq),
            "lt" => Ok(Self::Lt),
            "gt" => Ok(Self::Gt),
            "le" => Ok(Self::Le),
            "ge" => Ok(Self::Ge),
            "add" => Ok(Self::Add),
            "sub" => Ok(Self::Sub),
            "div" => Ok(Self::Div),
            "mod" => Ok(Self::Mod),
            "mul" => Ok(Self::Mul),
            "shr" => Ok(Self::Shr),
            "shl" => Ok(Self::Shl),
            "band" => Ok(Self::Band),
            "bor" => Ok(Self::Bor),
            "bnot" => Ok(Self::Bnot),
            "and" => Ok(Self::And),
            "or" => Ok(Self::Or),
            "xor" => Ok(Self::Xor),
            "inc" => Ok(Self::Inc),
            "dec" => Ok(Self::Dec),
            "ref" => Ok(Self::Ref),
            "struct" => Ok(Self::Struct),
            "enum" => Ok(Self::Enum),
            "export" => Ok(Self::Export),
            "else" => Ok(Self::Else),
            "call" => Ok(Self::Call),
            "return" => Ok(Self::Return),
            "structinit" => Ok(Self::StructInit),
            _ => bail!("Unknown keyword: {}", value)
        }
    }
}

impl TryFrom<String> for KeywordType {
    type Error = anyhow::Error;
    fn try_from(value: String) -> Result<Self, Self::Error> {
        value.as_str().try_into()
    }
}


