use std::io::BufRead;

use anyhow::bail;
use camino::Utf8Path;

use crate::{common::Loc, token::{KeywordType, Token, TokenType}};




pub struct Tokeniser {
    loc: Loc,
    tokens: Vec<Token>,
}

impl Tokeniser {
    pub fn new() -> Self {
        Self {
            loc: Loc::new(""),
            tokens: Vec::new(),
        }
    }

    pub fn tokenise(&mut self, file: &Utf8Path) -> anyhow::Result<()>{
        self.reset();
        self.loc.file = file.into();
        let Ok(f) = std::fs::File::open(file) else {
            log::error!("Failed to open file {file}, permissions?");
            bail!("");
        };
        let mut f = std::io::BufReader::new(f);

        let mut line = String::new();
        loop {
            match f.read_line(&mut line) {
                Ok(c) if c == 0 => break,
                Ok(_) => (),
                Err(e) => {
                    log::error!("Failed to read line from file: {e}");
                    bail!("");
                }
            }
            let mut chars = line.chars().into_iter().peekable();
            'line: while let Some(c) = chars.next() {
                match c {
                    '(' => self.tokens.push(Token::new(TokenType::ParenL, self.loc.clone())),
                    ')' => self.tokens.push(Token::new(TokenType::ParenR, self.loc.clone())),
                    '.' => self.tokens.push(Token::new(TokenType::Dot, self.loc.clone())),
                    ';' => break 'line,
                    '-' | '+' | '0'..='9' => {
                        let loc = self.loc.clone();
                        let mut buf = String::from(c);
                        let mut is_float = false;
                        while let Some(c) = chars.peek() {
                            match c {
                                '0'..='9' => {
                                    buf.push(*c);
                                    chars.next();
                                }
                                '.' => {
                                    if is_float {
                                        log::error!("{}: Invalid float", loc);
                                        bail!("");
                                    } else {
                                        is_float = true;
                                        buf.push(*c);
                                        chars.next();
                                    }
                                }
                                'x' | 'b' | 'o' => {
                                    if buf.len() != 1 {
                                        log::error!("{}: Invalid int", loc);
                                        bail!("");
                                    }
                                    buf.push(*c);
                                    chars.next();
                                }
                                _ => break
                            }
                            self.loc.col += 1;
                        }
                        if is_float {
                            let v;
                            match parse_int::parse::<f64>(&buf) {
                                Ok(f) => v = f,
                                Err(e) => {
                                    log::error!("{}: Failed to parse float: {e}", loc);
                                    bail!("");
                                }
                            }

                            self.tokens.push(Token::new(TokenType::Float(v), loc))
                        } else {
                            let v;
                            match parse_int::parse::<i64>(&buf) {
                                Ok(i) => v = i,
                                Err(e) => {
                                    log::error!("{}: Failed to parse float: {e}", loc);
                                    bail!("");
                                }
                            }

                            self.tokens.push(Token::new(TokenType::Int(v), loc))
                        } 
                    }
                    'a'..='z' | 'A'..='Z' | '_' => {
                        let loc = self.loc.clone();
                        let mut buf = String::from(c);
                        while let Some(c) = chars.peek() {
                            if !matches!(c, 'a'..='z' | 'A'..='Z' | '_' | '0'..='9') {
                                self.loc.col -= 1;
                                break;
                            }
                            buf.push(*c);
                            chars.next();
                            self.loc.col += 1;
                        }

                        if buf == "_" {
                            self.tokens.push(Token::new(TokenType::Underscore, self.loc.clone()));
                        } else if let Ok(kw) = KeywordType::try_from(buf.clone()) {
                            self.tokens.push(Token::new(TokenType::Keyword(kw), loc));
                        } else {
                            self.tokens.push(Token::new(TokenType::Ident(buf), loc));
                        }
                    }
                    '"' => {
                        let loc = self.loc.clone();
                        let mut buf = String::new();
                        let mut escaped = false;
                        while let Some(c) = chars.next() {
                            match c {
                                '"'  if escaped => {
                                    buf.push(c);
                                    escaped = false;
                                }
                                '"'  if !escaped => break,
                                '\\' if !escaped => escaped = true,
                                '\\' if escaped => {
                                    buf.push(c);
                                    escaped = false;
                                }
                                'n' if escaped => {
                                    buf.push('\n');
                                    escaped = false;
                                }
                                'r' if escaped => {
                                    buf.push('\r');
                                    escaped = false;
                                }
                                _ if !escaped => buf.push(c),
                                _ if escaped => {
                                    log::warn!("{}: Unknown escape \\{}, skipping and putting it as literal", self.loc, c);
                                    buf.push('\\');
                                    buf.push(c);
                                }
                                _ => unreachable!()
                            }
                            self.loc.col += 1;
                        }
                        self.tokens.push(Token::new(TokenType::String(buf), loc));
                    }
                    '\'' => {
                        let loc = self.loc.clone();
                        let mut buf = String::new();
                        let mut escaped = false;
                        while let Some(c) = chars.next() {
                            match c {
                                '\''  if escaped => {
                                    buf.push(c);
                                    escaped = false;
                                }
                                '\'' if !escaped => break,
                                '\\' if !escaped => escaped = true,
                                '\\' if escaped => {
                                    buf.push(c);
                                    escaped = false;
                                }
                                'n' if escaped => {
                                    buf.push('\n');
                                    escaped = false;
                                }
                                'r' if escaped => {
                                    buf.push('\r');
                                    escaped = false;
                                }
                                _ if !escaped => buf.push(c),
                                _ if escaped => {
                                    log::warn!("{}: Unknown escape \\{}, skipping and putting it as literal", self.loc, c);
                                    buf.push('\\');
                                    buf.push(c);
                                }
                                _ => unreachable!()
                            }
                            self.loc.col += 1;
                        }

                        if buf.len() > 1 {
                            log::error!("{}: Chars can only have 1 character", loc);
                            bail!("");
                        }
                        self.tokens.push(Token::new(TokenType::Char(buf.chars().nth(0).unwrap()), loc))
                    }
                    _ => ()
                }
                self.loc.col += 1;
            }
            self.loc.line += 1;
            self.loc.col = 1;
            line.clear();
        }

        Ok(())
    }

    fn reset(&mut self) {
        self.loc.col = 1;
        self.loc.line = 1;
        // self.loc.file = "".into();
        self.tokens.clear();
    }

    pub fn tokens(&self) -> &Vec<Token> {
        &self.tokens
    }
}
